///////////////////////////////////////////////////////////////////////////////
// Title      : DIOT WIC testbench top
// Project    : Distributed I/O Tier
///////////////////////////////////////////////////////////////////////////////
// File       : bases.vhd
// Author     : Denia Bouhired <denia.bouhired@cern.ch>
// Company    : CERN (BE-CO-HT)
// Standard   : VHDL
///////////////////////////////////////////////////////////////////////////////
// Description:
// Generic base classes for DIOT/WIC testbench
//
///////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2018 CERN
//
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 0.51 (the “License”) (which enables you, at your option,
// to treat this file as licensed under the Apache License 2.0); you may not
// use this file except in compliance with the License. You may obtain a copy
// of the License at http://solderpad.org/licenses/SHL-0.51.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// “AS IS” BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//
///////////////////////////////////////////////////////////////////////////////



`ifndef BASES__SV
`define BASES__SV


typedef class DIOT_stimulus_gen; //forward type definition
typedef t_dir;
class DIOT_trans; // parent class

    const DIOT_stimulus_gen owner;
    t_dir dir;
    byte unsigned payload [0:123];

    function new(DIOT_stimulus_gen owner = null); 
        this.owner = owner;
        this.payload='{124{'h00}};
        $write("New transaction created ...");
 endfunction : new


endclass

//typedef class WIC_trans; 

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~//
//
// Channel - FIFO channel that components can use to communicate.
// Passes objects of Transaction (base class) type, therefore
// can also pass objects of any derived class.


class Channel #(type Tr = DIOT_trans);


  // PUBLIC METHODS - This is how you use the channel:

  // Constructor: bound=0 for unbounded, bound=1 for single-place FIFO, etc
  extern function     new(int _bound = 0);

  // Push data onto the FIFO:
  // ~~~~~~~~~~~~~~~~~~~~~~~~
  // put: block until space free, then push the data
  extern task         put(Tr t);
  // try_put: if no free space, return 0 immediately;
  //          if some free space, push data and return 1 immediately
  extern function bit try_put(Tr t);

  // Pop data off the FIFO:
  // ~~~~~~~~~~~~~~~~~~~~~~
  // get: block until data available, then pop the data into (t)
  extern task         get(output Tr t);
  // try_get: if no data, return 0 immediately;
  //          if data available, pop data into (t) and return 1
  extern function bit try_get(output Tr t);

  // Peek: like get, but leave the data on the FIFO (don't pop)
  // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  // peek: block until data available, then return data in (t)
  extern task         peek(output Tr t);
  // try_peek: if no data, return 0 immediately;
  //           if data available, pass data in (t) and return 1
  extern function bit try_peek(output Tr t);

  // PRIVATE DATA MEMBERS AND METHODS OF THE CHANNEL CLASS
  // Users cannot touch these

  local Tr fifo[$];
  local event e_get, e_put;
  local int bound;

  extern local function        Tr pop();
  extern local function        void push(Tr t);
  extern local function         bit empty();
  extern local function         bit full();
  extern local task                 await_not_full();
  extern local task                 await_not_empty();

endclass : Channel



//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~//
// IMPLEMENTATIONS
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~//

function Channel::Tr Channel::pop();
  -> e_get;
  return fifo.pop_front();
endfunction : pop

function void Channel::push(Tr t);
  fifo.push_back(t);
  -> e_put;
endfunction : push

function bit Channel::empty();
  return (fifo.size() == 0);
endfunction : empty

function bit Channel::full();
  return (bound && (fifo.size() >= bound));
endfunction : full

task Channel::await_not_empty();
  while (empty()) @e_put;
endtask : await_not_empty

task Channel::await_not_full();
  while (full()) @e_get;
endtask : await_not_full

function Channel::new(int _bound = 0);
  bound = _bound;
endfunction : new

task Channel::put(Tr t);
  await_not_full();
  push(t);
endtask : put

task Channel::get(output Tr t);
  await_not_empty();
  t = pop();
endtask : get

function bit Channel::try_get(output Tr t);
  if (empty()) return 0;
  t = pop();
  return 1;
endfunction : try_get

function bit Channel::try_put(Tr t);
  if (full()) return 0;
  push(t);
  return 1;
endfunction : try_put

task Channel::peek(output Tr t);
  await_not_empty();
  t = fifo[0];
endtask : peek

function bit Channel::try_peek(output Tr t);
  if (empty()) return 0;
  t = fifo[0];
  return 1;
endfunction : try_peek

`endif
