///////////////////////////////////////////////////////////////////////////////
// Title      : DIOT WIC testbench top
// Project    : Distributed I/O Tier
///////////////////////////////////////////////////////////////////////////////
// File       : wic_classes.vhd
// Author     : Denia Bouhired <denia.bouhired@cern.ch>
// Company    : CERN (BE-CO-HT)
// Standard   : VHDL
///////////////////////////////////////////////////////////////////////////////
// Description:
// WIC specific base classes for DIOT/WIC testbench
//
///////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2018 CERN
//
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 0.51 (the “License”) (which enables you, at your option,
// to treat this file as licensed under the Apache License 2.0); you may not
// use this file except in compliance with the License. You may obtain a copy
// of the License at http://solderpad.org/licenses/SHL-0.51.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// “AS IS” BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//
///////////////////////////////////////////////////////////////////////////////



// Generic DIOT transaction class
// Targets any board

`include "serializable.svh"


typedef enum bit {dir_read, dir_write} t_dir; // direction of transaction
typedef class DIOT_stimulus_gen; //forward type definition
typedef class DIOT_Monitor; //forward type definition
typedef Channel #(DIOT_trans) DIOT_Channel;

// Child class of DIOT_trans for Digital IO board 
// DIO transaction
class WIC_trans;// extends DIOT_trans;
    const DIOT_stimulus_gen owner;
    static int next_ID;
    rand int slot; //slot number of destination peripheral
    const int ID; //ID of transaction;
    rand byte unsigned  dio_data [1:0];
    rand t_dir dir;
    
//TODO make slot ID unique !
/////////////////////////////
    constraint legal_slot { 
    slot >= 1;
    slot <=4;
}


 function new(DIOT_stimulus_gen owner = null); 
// number of io channels may not be needed, replace instead with array of length 16
    // 16 bit value
    this.owner = owner;
    ID = next_ID++; 
    //this.payload='{124{'h00}};
    $write("New transaction created ...");
 endfunction : new

 function void print;
    string kind = (dir==dir_read) ? "READ" : "WRITE";
    //string status = (data_status==1) ?  "SENT" : "READ";
    $display ("%s Cycle - Slot %0d [#%0d]: ", kind, slot, ID);
    $display ("%h, %h, %h, %h", dio_data[3], dio_data[2], dio_data[1], dio_data[0]);
 endfunction : print

 function WIC_trans copy();
    copy = new this;
    return copy;
 endfunction : copy

endclass

//--------------- TESTBENCH EXTENSION---------------
// Define other transaction types depending on peripheral board used
// class other_trans DIOT_trans

//....

// endclass : other_trans
//--------------- TESTBENCH EXTENSION---------------

//====================================================
// Stimulus generator
//====================================================

class DIOT_stimulus_gen;

    DIOT_Channel chan;
    //int slot;
    WIC_trans dio_trans;
    DIOT_trans current_temp;
    WIC_trans history [8][$]; //queue of all generated objects
    byte unsigned DIO_payload [0:123] = '{124{'h00}};
    function new ();
       this.dio_trans=new(this);
       this.current_temp=new(this);
       this.current_temp=new(this);
    endfunction : new;

    function WIC_trans gen_item;
        dio_trans = new(this);
        //current_temp.payload=DIO_payload;
        //if (!current_temp.randomize())
        void'(dio_trans.randomize());
        history[dio_trans.slot].push_back(dio_trans);
        
        return dio_trans;

    endfunction : gen_item;

    task flush ();
        this.current_temp.payload='{124{'h00}};
    endtask : flush 
    task body(int N = 8 );
        repeat (N) begin
            this.gen_item();
            this.dio_trans.print;
            if (this.dio_trans.dir ==dir_write)
                begin
                    current_temp.dir=dir_write;
                    case (dio_trans.slot)
                        1: current_temp.payload[0:3]= '{dio_trans.dio_data[1], dio_trans.dio_data[0] , {1{'h00}}, {1{'h00}}};
                        2: current_temp.payload[4:7]= '{dio_trans.dio_data[1], dio_trans.dio_data[0] , {1{'h00}}, {1{'h00}}};
                        3: current_temp.payload[8:11]= '{dio_trans.dio_data[1], dio_trans.dio_data[0] , {1{'h00}}, {1{'h00}}};
                        4: current_temp.payload[12:15]= '{dio_trans.dio_data[1], dio_trans.dio_data[0] , {1{'h00}}, {1{'h00}}};
                        5: current_temp.payload[16:19]= '{dio_trans.dio_data[1], dio_trans.dio_data[0] , {1{'h00}}, {1{'h00}}};
                        6: current_temp.payload[20:23]= '{dio_trans.dio_data[1], dio_trans.dio_data[0] , {1{'h00}}, {1{'h00}}};
                        7: current_temp.payload[24:27]= '{dio_trans.dio_data[1], dio_trans.dio_data[0] , {1{'h00}}, {1{'h00}}};
                        8: current_temp.payload[28:31]= '{dio_trans.dio_data[1], dio_trans.dio_data[0] , {1{'h00}}, {1{'h00}}};
                    endcase
                end
            else begin
                current_temp.dir=dir_read;
              //  chan.put(current_temp);
            end
            DIO_payload=current_temp.payload;
        end
            chan.put(current_temp);
    endtask
    
endclass

//====================================================
// DRIVER
//====================================================


class driver;

    virtual tb_fd_hook hook_fd; //no modport so far
    DIOT_Channel chan;
    byte unsigned DIO_payload [0:123];

    function new (virtual tb_fd_hook hook_fd);//, 
                  //virtual tb_fd_hook hook_fd);
        this.hook_fd = hook_fd;
    endfunction : new


    task body;
       DIOT_trans tr;
       chan.get(tr);
       drive (tr);
    endtask : body

    task drive (DIOT_trans tr);

       string kind = (tr.dir==dir_read) ? "READ" : "WRITE";
           $display("Direction is %s", kind);
       if (tr.dir ==dir_write) begin
              hook_fd.payload=tr.payload;
              this.hook_fd.sendFIPframe (tr.payload);
              #50us;      // silence time
       end
       else begin
             this.hook_fd.getFIPframe ();
             #100us;      // silence time
             this.hook_fd.dispFIPframe();

       end
    endtask : drive


endclass :driver


//====================================================
// TOFIX Monitor
//====================================================

class DIOT_Monitor; 
    virtual nf_fmc_if hook_nf; //no modport so far
    virtual tb_fd_hook hook_fd; //no modport so far
       // DIOT_Channel sink;

    function new(virtual nf_fmc_if hook_nf, virtual tb_fd_hook hook_fd);
        this.hook_nf  = hook_nf;
        this.hook_fd  = hook_fd;
      //  sink = new(1); 
      endfunction 

    task body;
        bit ok;
        DIOT_trans current;
        current=new();
        forever begin
            if (hook_nf.nf_var3_acc) 
            begin
                hook_fd.storeFIPframe();
               // current.dir=dir_write;
                hook_fd.dispFIPframe();
                
            end 
               // $display ("rx idle is %b No new data received!",hook_fd.getRxStatus() );
    end 
    // end

       // ok=sink.try_put(current.copy());
       // assert(ok);
    endtask : body;
endclass : DIOT_Monitor



//====================================================
// TODO Checker 
// This is a skeleton
//====================================================

class DIOT_Checker;    virtual nf_fmc_if hook_nf; //no modport so far
    // DIOT_Channel source;

  function new(virtual nf_fmc_if hook_nf);
        this.hook_nf  = hook_nf;
        // source = new(1); 
  endfunction 

task body; begin
       DIOT_trans tr;
       // source.get(tr);
end

    // ok=source.try_put(current.copy());
    // assert(ok);
endtask : body;
endclass : DIOT_Checker


//====================================================
// TB_ENV
//====================================================

class tb_env;
    int num_slots;
    virtual nf_fmc_if hook_nf; //no modport so far
    virtual tb_fd_hook hook_fd; //no modport so far
    DIOT_stimulus_gen gen;
    driver bfm;
    DIOT_Monitor mon;
    DIOT_Channel chan;
    DIOT_Channel mon_chan;
    function new(int num_slots, virtual nf_fmc_if hook_nf, virtual tb_fd_hook hook_fd);
        this.num_slots = num_slots;
        this.gen = new();
        this.bfm = new(hook_fd);
        this.mon = new(hook_nf, hook_fd);
        this.chan = new(1);
        this.mon_chan = new(1);
        this.gen.chan= this.chan;
        this.bfm.chan= this.chan;
    endfunction : new;

    task run (int N = 20);
        //repeat (N)
           // begin
           gen.flush;
           fork
                gen.body(N);
                bfm.body;
                // mon.body;
           join_any
                $display("FIP frame has been pushed...");
    endtask : run ;
 
endclass



