/*
 * Copyright (C) 2018 CERN
 * Author: Maciej Suminski <maciej.suminski@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>

#include "libmfipp.h"
#include "nanofip.h"

constexpr int VAR_SIZE = 124;

int main(int argc, char *argv[])
{
    auto mfip = MasterFIP::OpenByLUN(0);

    // Hardware and software config for MasterFIP
    mstrfip_hw_cfg hw_cfg = {0};
    hw_cfg.enable_ext_trig = 0;
    hw_cfg.enable_int_trig = 0;
    hw_cfg.enable_ext_trig_term = 0;
    hw_cfg.turn_around_ustime = 14;         // 14 us corresponds to 1 Mbit/s

    mstrfip_sw_cfg sw_cfg = {0};
    sw_cfg.irq_thread_prio = 50;
    sw_cfg.diag_thread_prio = 20;

    // Cycle configuration (60 ms duration)
    CycleSimple cycle(*mfip, 60000);

    // Agents configuration
    nanoFIP agent1(0x01, VAR_SIZE);
    agent1.AddVariables(cycle.GetPeriodicVarWindow());

    // Variable representing I/O state
    uint8_t prod = 0;

    // Initialization
    for(int i = 0; i < VAR_SIZE; ++i) {
        agent1.SetProdVarData(i, 0);
    }

    // Produced variable handler
    agent1.SetProdVarCb([&](nanoFIP& nf) {
        ++prod;

        for(int i = 0; i < VAR_SIZE; ++i) {
            agent1.SetProdVarData(i, prod);
        }

        std::cout << "variable produced = " << (int) prod << std::endl;
    } );

    agent1.SetConsVarCb([&](nanoFIP& nf) {
        uint8_t cons = nf.GetConsVarData(0);

        for(int i = 1; i < VAR_SIZE; ++i) {
            uint32_t var_data = agent1.GetConsVarData(i);

            if(cons != var_data) {
                std::cout << "consumed variable differs at index " << i
                          << " (" << (int) cons << " != " << (int) var_data << ")" << std::endl;
            }
        }

        std::cout << "variable consumed = " << (int) cons << std::endl;
    } );

    mfip->Configure(hw_cfg, sw_cfg);
    mfip->StartCycle(cycle);

    // Keep running until a signal arrives
    WaitForSignal();

    //mfip->StopCycle();

    return 0;
}
